@@ -10,13 +10,24 @@ from typing import Any, Dict, List, Tuple, Union
 import requests
 
 from loguru import logger
+from pydantic import parse_obj_as
 
 from aoe2netwrapper.exceptions import Aoe2NetException
+from aoe2netwrapper.models import (
+    LastMatchResponse,
+    LeaderBoardResponse,
+    MatchLobby,
+    NumOnlineResponse,
+    RatingTimePoint,
+    StringsResponse,
+)
 
 
 class AoE2NetAPI:
     """
-    The 'AoE2NetAPI' class is a client that encompasses the https://aoe2.net/#api API  functions.
+    The 'AoE2NetAPI' class is a client that encompasses the https://aoe2.net/#api API endpoints.
+    Each method in this class corresponds name for name to an endpoint, and will do the work in querying then
+    parsing and validating the response before returning it.
     """
 
     API_BASE_URL: str = "https://aoe2.net/api"
@@ -38,7 +49,7 @@ class AoE2NetAPI:
     def __repr__(self) -> str:
         return f"Client for <{self.API_BASE_URL}>"
 
-    def strings(self, game: str = "aoe2de") -> dict:
+    def strings(self, game: str = "aoe2de") -> StringsResponse:
         """
         Requests a list of strings used by the API.
 
@@ -48,18 +59,19 @@ class AoE2NetAPI:
                 Empires 2: Definitive Edition).
 
         Returns:
-            A dictionnary with different strings used by the API, each with their list of possible
-            values (also a dict each).
+            A StringsResponse validated object encapsulating the strings used by the API.
         """
         logger.debug("Preparing parameters for strings query")
         query_params = {"game": game}
 
-        return _get_request_response_json(
+        processed_response = _get_request_response_json(
             session=self.session,
             url=self.STRINGS_ENDPOINT,
             params=query_params,
             timeout=self.timeout,
         )
+        logger.trace(f"Validating response from '{self.STRINGS_ENDPOINT}'")
+        return StringsResponse(**processed_response)
 
     def leaderboard(
         self,
@@ -70,7 +82,7 @@ class AoE2NetAPI:
         search: str = None,
         steam_id: int = None,
         profile_id: int = None,
-    ) -> dict:
+    ) -> LeaderBoardResponse:
         """
         Request the current leaderboards.
 
@@ -85,18 +97,19 @@ class AoE2NetAPI:
                 Defaults to 1.
             count (int): Number of leaderboard entries to get (warning: must be 10000 or less).
                 Defaults to 10.
-            search (str): To perform the search for a specific player, from their name.
-            steam_id (int): To perform the search for a specific player, from their steamID64
-                (ex: 76561199003184910).
-            profile_id (int): To perform the search for a specific player, from their profile ID
-                (ex: 459658).
+            search (str): Optional. To perform the search for a specific player, from their name.
+            steam_id (int): Optional. To perform the search for a specific player, from their
+                steamID64 (ex: 76561199003184910).
+            profile_id (int): Optional. To perform the search for a specific player, from their
+                profile ID (ex: 459658).
 
         Raises:
             Aoe2NetException: if the 'count' parameter exceeds 10 000.
 
         Returns:
-            A dictionnary with the different parameters used for the query, the 'total' amount of
-            hits, and the leaderboard as a list profile entries for each ranking (a dict each).
+            A LeaderBoardResponse validated object with the different parameters used for the
+            query, the total amount of hits, and the leaderboard as a list profile entries for
+            each ranking.
         """
         if count > 10_000:
             logger.error(f"'count' has to be 10000 or less, but {count} was provided.")
@@ -113,14 +126,16 @@ class AoE2NetAPI:
             "profile_id": profile_id,
         }
 
-        return _get_request_response_json(
+        processed_response = _get_request_response_json(
             session=self.session,
             url=self.LEADERBOARD_ENDPOINT,
             params=query_params,
             timeout=self.timeout,
         )
+        logger.trace(f"Validating response from '{self.LEADERBOARD_ENDPOINT}'")
+        return LeaderBoardResponse(**processed_response)
 
-    def lobbies(self, game: str = "aoe2de") -> List[dict]:
+    def lobbies(self, game: str = "aoe2de") -> List[MatchLobby]:
         """
         Request all open lobbies.
 
@@ -130,21 +145,24 @@ class AoE2NetAPI:
                 Empires 2: Definitive Edition).
 
         Returns:
-            A list of dictionaries, each one encapsulating the data for a currently open lobby.
+            A list of MatchLobby valideted objects, each one encapsulating the data for a currently
+            open lobby.
         """
         logger.debug("Preparing parameters for open lobbies query")
         query_params = {"game": game}
 
-        return _get_request_response_json(
+        processed_response = _get_request_response_json(
             session=self.session,
             url=self.LOBBIES_ENDPOINT,
             params=query_params,
             timeout=self.timeout,
         )
+        logger.trace(f"Validating response from '{self.LOBBIES_ENDPOINT}'")
+        return parse_obj_as(List[MatchLobby], processed_response)
 
     def last_match(
         self, game: str = "aoe2de", steam_id: int = None, profile_id: int = None
-    ) -> dict:
+    ) -> LastMatchResponse:
         """
         Request the last match the player started playing, this will be the current match if they
         are still in game. Either 'steam_id' or 'profile_id' required.
@@ -160,25 +178,25 @@ class AoE2NetAPI:
             Aoe2NetException: if the not one of 'steam_id' or 'profile_id' are provided.
 
         Returns:
-            A dictionary with the information of the game, including the following keys:
-            'profile_id', 'steam_id', 'name', 'clan', 'country' & 'last_match', with 'last_match'
-            being a dictionary with the match's information.
+            A LastMatchResponse validated object with the information of the game, including the
+            following attributes: 'profile_id', 'steam_id', 'name', 'clan', 'country' and
+            'last_match'.
         """
         if not steam_id and not profile_id:
             logger.error("Missing one of 'steam_id', 'profile_id'.")
-            raise Aoe2NetException(
-                "Either 'steam_id' or 'profile_id' required, please provide one."
-            )
+            raise Aoe2NetException("Either 'steam_id' or 'profile_id' required, please provide one.")
 
         logger.debug("Preparing parameters for last match query")
         query_params = {"game": game, "steam_id": steam_id, "profile_id": profile_id}
 
-        return _get_request_response_json(
+        processed_response = _get_request_response_json(
             session=self.session,
             url=self.LAST_MATCH_ENDPOINT,
             params=query_params,
             timeout=self.timeout,
         )
+        logger.trace(f"Validating response from '{self.LAST_MATCH_ENDPOINT}'")
+        return LastMatchResponse(**processed_response)
 
     def match_history(
         self,
@@ -187,7 +205,7 @@ class AoE2NetAPI:
         count: int = 10,
         steam_id: int = None,
         profile_id: int = None,
-    ) -> List[dict]:
+    ) -> List[MatchLobby]:
         """
         Request the match history for a player. Either 'steam_id' or 'profile_id' required.
 
@@ -205,8 +223,8 @@ class AoE2NetAPI:
             Aoe2NetException: if the not one of 'steam_id' or 'profile_id' are provided.
 
         Returns:
-            A list of dictionaries, each one encapsulating the data for one of the player's
-            previous matches.
+            A list of MatchLobby validated objects, each one encapsulating the data for one of the
+            player's previous matches.
         """
         if count > 1_000:
             logger.error(f"'count' has to be 1000 or less, but {count} was provided.")
@@ -214,9 +232,7 @@ class AoE2NetAPI:
 
         if not steam_id and not profile_id:
             logger.error("Missing one of 'steam_id', 'profile_id'.")
-            raise Aoe2NetException(
-                "Either 'steam_id' or 'profile_id' required, please provide one."
-            )
+            raise Aoe2NetException("Either 'steam_id' or 'profile_id' required, please provide one.")
 
         logger.debug("Preparing parameters for match history query")
         query_params = {
@@ -227,22 +243,24 @@ class AoE2NetAPI:
             "profile_id": profile_id,
         }
 
-        return _get_request_response_json(
+        processed_response = _get_request_response_json(
             session=self.session,
             url=self.MATCH_HISTORY_ENDPOINT,
             params=query_params,
             timeout=self.timeout,
         )
+        logger.trace(f"Validating response from '{self.MATCH_HISTORY_ENDPOINT}'")
+        return parse_obj_as(List[MatchLobby], processed_response)
 
     def rating_history(
         self,
         game: str = "aoe2de",
         leaderboard_id: int = 3,
         start: int = 0,
-        count: int = 100,
+        count: int = 20,
         steam_id: int = None,
         profile_id: int = None,
-    ) -> List[dict]:
+    ) -> List[RatingTimePoint]:
         """
         Requests the rating history for a player. Either 'steam_id' or 'profile_id' required.
 
@@ -264,9 +282,9 @@ class AoE2NetAPI:
             Aoe2NetException: if the not one of 'steam_id' or 'profile_id' are provided.
 
         Returns:
-            A list of dictionaries, each one encapsulating data at a certain point in time
-            corresponding to a match played by the player, including the rating, timestamp of the
-            match, streak etc.
+            A list of RatingTimePoint validated objects, each one encapsulating data at a certain
+            point in time corresponding to a match played by the player, including the rating,
+            timestamp of the match, streaks etc.
         """
         if count > 10_000:
             logger.error(f"'count' has to be 10 000 or less, but {count} was provided.")
@@ -274,9 +292,7 @@ class AoE2NetAPI:
 
         if not steam_id and not profile_id:
             logger.error("Missing one of 'steam_id', 'profile_id'.")
-            raise Aoe2NetException(
-                "Either 'steam_id' or 'profile_id' required, please provide one."
-            )
+            raise Aoe2NetException("Either 'steam_id' or 'profile_id' required, please provide one.")
 
         logger.debug("Preparing parameters for rating history query")
         query_params = {
@@ -288,14 +304,16 @@ class AoE2NetAPI:
             "profile_id": profile_id,
         }
 
-        return _get_request_response_json(
+        processed_response = _get_request_response_json(
             session=self.session,
             url=self.RATING_HISTORY_ENDPOINT,
             params=query_params,
             timeout=self.timeout,
         )
+        logger.trace(f"Validating response from '{self.RATING_HISTORY_ENDPOINT}'")
+        return parse_obj_as(List[RatingTimePoint], processed_response)
 
-    def matches(self, game: str = "aoe2de", count: int = 10, since: int = None) -> List[dict]:
+    def matches(self, game: str = "aoe2de", count: int = 10, since: int = None) -> List[MatchLobby]:
         """
         Request matches after a specific time: the match history in an optionally given time
         window.
@@ -314,8 +332,8 @@ class AoE2NetAPI:
             Aoe2NetException: if the 'count' parameter exceeds 1000.
 
         Returns:
-            A list of dictionaries, each one encapsulating the data for one of the played matches
-            during the time window queried for.
+            A list of MatchLobby validated objects, each one encapsulating the data for one of the
+            played matches during the time window queried for.
         """
         if count > 1000:
             logger.error(f"'count' has to be 1000 or less, but {count} was provided.")
@@ -328,14 +346,16 @@ class AoE2NetAPI:
             "since": since,
         }
 
-        return _get_request_response_json(
+        processed_response = _get_request_response_json(
             session=self.session,
             url=self.MATCHES_ENDPOINT,
             params=query_params,
             timeout=self.timeout,
         )
+        logger.trace(f"Validating response from '{self.MATCHES_ENDPOINT}'")
+        return parse_obj_as(List[MatchLobby], processed_response)
 
-    def match(self, game: str = "aoe2de", uuid: str = None, match_id: int = None) -> dict:
+    def match(self, game: str = "aoe2de", uuid: str = None, match_id: int = None) -> MatchLobby:
         """
         Request details about a match. Either 'uuid' or 'match_id' required.
 
@@ -350,7 +370,7 @@ class AoE2NetAPI:
             Aoe2NetException: if the not one of 'uuid' or 'match_id' are provided.
 
         Returns:
-            A dictionary with the information of the specific match, including.
+            A MatchLobby validated object with the information of the specific match, including.
         """
         if not uuid and not match_id:
             logger.error("Missing one of 'uuid', 'match_id'.")
@@ -363,14 +383,16 @@ class AoE2NetAPI:
             "match_id": match_id,
         }
 
-        return _get_request_response_json(
+        processed_response = _get_request_response_json(
             session=self.session,
             url=self.MATCH_ENDPOINT,
             params=query_params,
             timeout=self.timeout,
         )
+        logger.trace(f"Validating response from '{self.MATCH_ENDPOINT}'")
+        return MatchLobby(**processed_response)
 
-    def num_online(self, game: str = "aoe2de") -> dict:
+    def num_online(self, game: str = "aoe2de") -> NumOnlineResponse:
         """
         Number of players in game and an estimate of the number current playing multiplayer.
 
@@ -380,19 +402,21 @@ class AoE2NetAPI:
                 Empires 2: Definitive Edition).
 
         Returns:
-            A dictionary with the app id and a list of the estimated 'num_players' metrics at
-            different timestamps ('steam', 'multiplayer', 'looking', 'in_game', 'multiplayer_1h' &
-            'multiplayer_24h').
+            A NumOnlineResponse validated object with the app id and a list of PlayerCountTimePoint
+            validated objects encapsulating estimated metrics at different timestamps ('steam',
+            'multiplayer', 'looking', 'in_game', 'multiplayer_1h' & 'multiplayer_24h').
         """
         logger.debug("Preparing parameters for number of online players query")
         query_params = {"game": game}
 
-        return _get_request_response_json(
+        processed_response = _get_request_response_json(
             session=self.session,
             url=self.NUMBER_ONLINE_ENDPOINT,
             params=query_params,
             timeout=self.timeout,
         )
+        logger.trace(f"Validating response from '{self.NUMBER_ONLINE_ENDPOINT}'")
+        return NumOnlineResponse(**processed_response)
 
 
 # ----- Helpers ----- #
@@ -425,8 +449,6 @@ def _get_request_response_json(
 
     response = session.get(url, params=params, headers=default_headers, timeout=timeout)
     if response.status_code != 200:
-        logger.error(
-            f"GET request at '{response.url}' returned a {response.status_code} status code"
-        )
+        logger.error(f"GET request at '{response.url}' returned a {response.status_code} status code")
         raise Aoe2NetException(f"Expected status code 200 - got {response.status_code} instead.")
     return response.json()
